import struct
from enum import Enum


class BasicType(Enum):
    String = "String"
    Int = "Int"
    Float = "Float"
    ByteArray = "ByteArray"
    List = "List"
    Map = "Map"


size_of_uint64 = struct.calcsize("<Q")


class RecordType:
    def __init__(self, basic_type):
        self._basicType = basic_type
        if basic_type == BasicType.Map:
            self._keys = []
            self._properties = {}
        elif basic_type == BasicType.List:
            self._item = None

    @classmethod
    def from_define(cls, path, define):
        if isinstance(define, dict):
            result = RecordType(BasicType.Map)
            for key, value in define.items():
                result._keys.append(key)
                result._properties[key] = cls.from_define(path + [key], value)
            return result
        elif isinstance(define, list):
            if len(define) != 1:
                raise Exception("Bad length", define, '.'.join(path))
            result = RecordType(BasicType.List)
            result._item = cls.from_define(path + ['item'], define[0])
            return result
        elif isinstance(define, BasicType):
            return RecordType(define)
        else:
            raise Exception("Invalid type", define, '.'.join(path))

    def dump(self):
        result = {
            "BasicType": self._basicType.value,
        }
        if self._basicType == BasicType.Map:
            result["Keys"] = self._keys
            result["Properties"] = dict(zip(self._properties, map(RecordType.dump, self._properties.values())))
        elif self._basicType == BasicType.List:
            result["Item"] = self._item.dump()

        return result

    @classmethod
    def load(cls, dump):
        result = RecordType(BasicType(dump["BasicType"]))
        if dump["BasicType"] == "Map":
            result._keys = dump["Keys"]
            result._properties = dict(zip(dump["Properties"], map(RecordType.load, dump["Properties"].values())))
        elif dump["BasicType"] == "List":
            result._item = RecordType.load(dump["Item"])
        return result

    def encode(self, data, buffer):
        if self._basicType == BasicType.Map:
            assert isinstance(data, dict), "but get: {}".format(data)
            length = len(self._keys)
            header_length = (length + 1) * size_of_uint64
            start = len(buffer)
            header = [0] * (length + 1)
            buffer.extend(bytearray(header_length))
            current = 0
            for i, key in enumerate(self._keys):
                header[i] = current
                property_type = self._properties[key]
                value = data[key]
                item_length = property_type.encode(value, buffer)
                current += item_length
            header[length] = current
            buffer[start:start + header_length] = struct.pack('<{}Q'.format(length + 1), *header)
            return header_length + current
        elif self._basicType == BasicType.List:
            assert isinstance(data, list), "but get: {}".format(data)
            length = len(data)
            header_length = (length + 2) * size_of_uint64
            start = len(buffer)
            header = [length] * (length + 2)
            buffer.extend(bytearray(header_length))
            current = 0
            for i, value in enumerate(data):
                header[i + 1] = current
                item_length = self._item.encode(value, buffer)
                current += item_length
            header[length + 1] = current
            buffer[start:start + header_length] = struct.pack('<{}Q'.format(length + 2), *header)
            return header_length + current
        elif self._basicType == BasicType.String:
            assert isinstance(data, str), "but get: {}".format(data)
            b = data.encode()
            buffer.extend(b)
            return len(b)
        elif self._basicType == BasicType.Int:
            assert isinstance(data, int), "but get: {}".format(data)
            b = struct.pack("<Q", data)
            buffer.extend(b)
            return len(b)
        elif self._basicType == BasicType.Float:
            assert isinstance(data, float), "but get: {}".format(data)
            b = struct.pack("<d", data)
            buffer.extend(b)
            return len(b)
        elif self._basicType == BasicType.ByteArray:
            assert (isinstance(data, bytes) or isinstance(data, bytearray)), "but get: {}".format(data)
            buffer.extend(data)
            return len(data)
        else:
            raise Exception("Should not run here")

    def decode(self, data):
        if self._basicType == BasicType.Map:
            length = len(self._keys)
            header_length = (length + 1) * size_of_uint64
            header = struct.unpack("<{}Q".format(int(header_length / size_of_uint64)), data[0:header_length].tobytes())
            result = {}
            current_pos = 0
            for i, key in enumerate(self._keys):
                next_pos = header[i + 1]
                d = data[header_length + current_pos: header_length + next_pos]
                result[key] = self._properties[key].decode(d)
                current_pos = next_pos
            return result
        elif self._basicType == BasicType.List:

            length = struct.unpack("<Q", data[0:8].tobytes())[0]
            header_length = (length + 2) * size_of_uint64
            header = struct.unpack("<{}Q".format(int(header_length / size_of_uint64)), data[0:header_length].tobytes())
            result = []
            current_pos = 0
            for i in range(length):
                next_pos = header[i + 2]
                d = data[header_length + current_pos:header_length + next_pos]
                result.append(self._item.decode(d))
                current_pos = next_pos

            return result
        elif self._basicType == BasicType.String:
            return data.tobytes().decode()
        elif self._basicType == BasicType.Int:
            return struct.unpack("<Q", data.tobytes())[0]
        elif self._basicType == BasicType.Float:
            return struct.unpack("<d", data.tobytes())[0]
        elif self._basicType == BasicType.ByteArray:
            return data.tobytes()
        else:
            raise Exception("Should not run here")


def new_record_type(define):
    return RecordType.from_define([], define)



from seeta_dataset.readonly_dataset import ReadOnlyDataset
from seeta_dataset.writeonly_dataset import WriteOnlyDataset


class DataCenter(object):
    def __init__(self, root):
        self._root = root

    def load_dataset(self, dataset_id):
        return ReadOnlyDataset(self._root, dataset_id)

    def create_dataset(self, record_type, dataset_id=None):
        return WriteOnlyDataset(self._root, dataset_id, record_type)



import json
import struct
import os
from seeta_dataset import base


class WriteOnlyDataset:
    """Create a new seeta dataset for writing seeta records
    Args:
        base_path: ...
        dataset_id: ...
        record_type: ...
    Examples:
    ```python
    >>> import seeta_dataset as sd
    >>> rt = sd.record_type.new_record_type({
    >>>        "s": BasicType.String,
    >>>        "i": BasicType.Int,
    >>>        "f": BasicType.Float,
    >>>        "l_s": [BasicType.String],
    >>>        "l_i": [BasicType.Int],
    >>>        "m": {
    >>>            "x": BasicType.Int,
    >>>            "y": BasicType.Float,
    >>>            "s": BasicType.String,
    >>>        },
    >>>        "l_m": [{
    >>>            "x": BasicType.Int,
    >>>            "y": BasicType.Float,
    >>>        }]
    >>>    })
    >>> with WriteOnlyDataset("./", "test", rt) as dataset:
    >>>     dataset.write({
    >>>         "s": "ssssss",
    >>>         "i": 1,
    >>>         "f": 1.1,
    >>>         "l_s": ["l_s_1", "l_s_2"],
    >>>         "l_i": [1, 2, 3],
    >>>         "m": {
    >>>             "x": 1,
    >>>             "y": 1.1,
    >>>             "s": "m.s",
    >>>         },
    >>>         "l_m": [{
    >>>             "x": 1,
    >>>             "y": 1.1,
    >>>             "s": "m.s",
    >>>         }, {
    >>>             "x": 1,
    >>>             "y": 1.1,
    >>>             "s": "m.s",
    >>>         }]
    >>>     })
    >>>
    ```
    """
    def __init__(self, base_path, dataset_id=None, record_type=None):
        self._base_path = base_path
        self._dataset_id = dataset_id or base.DEFAULT_DATA_NAME
        self._record_type = record_type

        self._version = 0

        meta_file_path = os.path.join(self._base_path, "{}.meta".format(self._dataset_id))
        if os.path.exists(meta_file_path):
            raise base.FileExistError(meta_file_path)
        self._meta_file = open(meta_file_path, 'w')

        index_file_path = os.path.join(self._base_path, "{}.index".format(self._dataset_id))
        if os.path.exists(index_file_path):
            raise base.FileExistError(index_file_path)
        self._index_file = open(index_file_path, 'wb')

        data_file_path = os.path.join(self._base_path, "{}.data".format(self._dataset_id))
        if os.path.exists(data_file_path):
            raise base.FileExistError(data_file_path)
        self._data_file = open(data_file_path, 'wb')

        self._indexes = []
        self._current = 0

    def get_id(self):
        return self._dataset_id

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            return None
        self.close()
        return True

    def close(self):
        self._meta_file.write(json.dumps(self.dump()))
        self._meta_file.close()
        self._indexes.append(self._current)
        index_data = struct.pack('<{}Q'.format(len(self._indexes)), *self._indexes)
        self._index_file.write(index_data)
        self._index_file.close()
        self._data_file.close()

    def dump(self):
        return {
            'Version': self._version,
            "ID": self._dataset_id,
            "RecordType": self._record_type.dump(),
        }

    def write(self, record):
        buffer = bytearray()
        length = self._record_type.encode(record, buffer)
        self._data_file.write(buffer)
        self._indexes.append(self._current)
        self._current += length


if __name__ == '__main__':
    from seeta_dataset.record_type import new_record_type, BasicType
    rt = new_record_type({
        "s": BasicType.String,
        "i": BasicType.Int,
        "f": BasicType.Float,
        "l_s": [BasicType.String],
        "l_i": [BasicType.Int],
        "m": {
            "x": BasicType.Int,
            "y": BasicType.Float,
            "s": BasicType.String,
        },
        "l_m": [{
            "x": BasicType.Int,
            "y": BasicType.Float,
        }],
        "b": BasicType.ByteArray,
    })
    with WriteOnlyDataset(".", "test", rt) as dataset:
        dataset.write({
            "s": "ssssss",
            "i": 1,
            "f": 1.1,
            "l_s": ["l_s_1", "l_s_2"],
            "l_i": [1, 2, 3],
            "m": {
                "x": 1,
                "y": 1.1,
                "s": "m.s",
            },
            "l_m": [{
                "x": 1,
                "y": 1.1,
                "s": "m.s",
            }, {
                "x": 1,
                "y": 1.1,
                "s": "m.s",
            }],
            "b": bytes(b'abcdefg')
        })

    print("done")



from setuptools import setup, find_packages


setup(name='seeta-dataset',
      version='0.0.7',
      platforms='any',
      maintainer_email='seetaas@seetatech.com',
      description='SeeTaaS DataSet',
      packages=find_packages(),
      install_requires=[
      ],
      classifiers=[
          'Programming Language :: Python',
          'Operating System :: OS Independent',
      ])



import json
import struct
from os import path
import random
from seeta_dataset.record_type import RecordType
from seeta_dataset import base
import glob


class ReadOnlyDataset:
    """Read seeta records from seeta dataset
    Args:
        base_path: ...
        dataset_id: ...
    Examples:
    ```python
    read seeta dataset in order
    >>> dataset = ReadOnlyDataset("./", "test")
    >>> for data in dataset.read():
    >>>     print(data)
    read seeta dataset randomly
    >>> dataset = ReadOnlyDataset("./", "test")
    >>> for data in dataset.random_read():
    >>>     print(data)
    ```
    """
    def __init__(self, base_path, dataset_id):
        dataset_id = dataset_id or base.DEFAULT_DATA_NAME
        if not path.isfile(path.join(base_path, "{}.meta".format(dataset_id))):
            dataset_id = glob.glob("{}/*.data".format(base_path))[0].split(".")[0]
        self._dataset_id = dataset_id
        self._base_path = base_path

        meta_file_path = path.join(self._base_path, "{}.meta".format(self._dataset_id))
        with open(meta_file_path, 'r') as meta_file:
            meta_data = meta_file.read()
            meta_obj = json.loads(meta_data)
            self._version = meta_obj["Version"]
            self._record_type = RecordType.load(meta_obj["RecordType"])

        index_file_path = path.join(self._base_path, "{}.index".format(self._dataset_id))
        with open(index_file_path, 'rb') as index_file:
            index_data = index_file.read()
            self._indexes = struct.unpack("<{}Q".format(int(len(index_data)/struct.calcsize("Q"))), index_data)

        data_file_path = path.join(self._base_path, "{}.data".format(self._dataset_id))
        self._data_file = open(data_file_path, 'rb')
        self._cursor = 0

    def read(self, start=0, count=None):
        if count is None:
            count = self.record_count() - start
        current_pos = self._indexes[start]
        self._data_file.seek(current_pos, 0)
        for i in range(count):
            next_pos = self._indexes[start+i+1]
            d = self._data_file.read(next_pos-current_pos)
            yield self._record_type.decode(memoryview(d))
            current_pos = next_pos

    def random_read(self):
        start_index = list(range(len(self._indexes) - 1))
        random.shuffle(start_index)
        for si in start_index:
            start_pos = self._indexes[si]
            end_pos = self._indexes[si + 1]
            self._data_file.seek(start_pos, 0)
            d = self._data_file.read(end_pos - start_pos)
            yield self._record_type.decode(memoryview(d))

    def record_count(self):
        return len(self._indexes)-1

    def data_size(self):
        return self._indexes[-1]

    def get(self):
        start_pos = self._indexes[self._cursor]
        end_pos = self._indexes[self._cursor + 1]
        d = self._data_file.read(end_pos - start_pos)
        self._cursor += 1
        return self._record_type.decode(memoryview(d))

    def set_cursor(self, index):
        start = self._indexes[index]
        self._data_file.seek(start, 0)
        self._cursor = index

    def close(self):
        self._data_file.close()


if __name__ == "__main__":
    dataset = ReadOnlyDataset(".", "test")
    for data in dataset.read():
        print(data)
